#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <mpi.h>

#define MASTER 0

typedef enum ArchType_ {UNKNOWN, LINUX, KSR, SUN4, SOLARIS} ArchType;
typedef enum CPUType_ {OTHERCPU, PPRO, P2KLAMATH, P2DESCHUTES} CPUType;

char *ArchName[] = {"Unknown", "LINUX", "ksr", "sun4", "solaris"};
char *CPUName[] = {"Unknown", "Pentium Pro", "Pentium II (Klamath)", "Pentium II (Deschutes)"};

typedef struct HostInfo_
{
	int      Rank;
	int      nTasks;
	ArchType OSID;
	CPUType  CPUID;
	char     Name[80];
	char     Machine[80];
	char     Arch[80];
	char     OS[80];
	char     CPU[80];
	char     OSversion[80];
	char     OSrelease[80];
}
HostInfo_t;

char *SysCommand(const char *command)
{
	FILE *fp;
	static char Buffer[256];
	int i;

	i = 0;
	fp = popen(command, "r");
	Buffer[i] = fgetc(fp);
	if (Buffer[i] != '\n') ++i;
	while (!feof(fp)) {
		Buffer[i] = fgetc(fp);
		if (Buffer[i] != '\n') ++i;
	}
	Buffer[i-1] = 0x0;
	pclose(fp);
	return Buffer;
}

void getarch(HostInfo_t *HostInfo)
{
	FILE *fp;
	char FileName[80];
	char Arch[80], cpu[80];

	strcpy(FileName, "/bin/uname");
	if ((fp = fopen(FileName, "r")) != NULL) {
		fclose(fp);
		sprintf(Arch, "%s", SysCommand("/bin/uname -s"));
	}

	if ((fp = fopen("/bin/uname", "r")) != NULL) {
		fclose(fp);
		sprintf(HostInfo->OSrelease, "%s", SysCommand("/bin/uname -r"));
		sprintf(HostInfo->OSversion, "%s", SysCommand("/bin/uname -v"));
		sprintf(HostInfo->Machine, "%s", SysCommand("/bin/uname -m"));
	}

	gethostname(HostInfo->Name, 79);

	if (!strcmp(Arch, "Linux")) {
		HostInfo->OSID = LINUX;
	} else if (!strcmp(Arch, "ksr1")) {
		HostInfo->OSID = KSR;
	} else if (!strcmp(Arch, "sun4")) {
		HostInfo->OSID = SUN4;
	} else if (!strcmp(Arch, "SunOS")) {
		if (HostInfo->OSrelease[0] == '5')
			HostInfo->OSID = SOLARIS;
		else
			HostInfo->OSID = SUN4;
	}

	strcpy(HostInfo->OS, ArchName[HostInfo->OSID]);

	sprintf(cpu, "%s", SysCommand("cat /proc/cpuinfo | grep \"model name\"| head -n 1 | cut -f 2 -d :"));
	if (strstr(cpu, " Pentium Pro")) {
		HostInfo->CPUID = PPRO;
	} else if (strstr(cpu, " Pentium II (Klamath)")) {
		HostInfo->CPUID = P2KLAMATH;
	} else if (strstr(cpu, " Pentium II (Deschutes)")) {
		HostInfo->CPUID = P2DESCHUTES;
	} else {
		HostInfo->CPUID = OTHERCPU;
	}

	strcpy(HostInfo->CPU, CPUName[HostInfo->CPUID]);

	MPI_Comm_rank(MPI_COMM_WORLD, &HostInfo->Rank);
	MPI_Comm_size(MPI_COMM_WORLD, &HostInfo->nTasks);
}

int main(int argc, char *argv[])
{
	int  *RankList;
	CPUType         *CPUIdList;
	char  **NameList;
	int  i;
	HostInfo_t HostInfo;

	MPI_Init(&argc, &argv);
	getarch(&HostInfo);

	printf("%d ", HostInfo.Rank);  /* Rank in MPI_COMM_WORLD */
	printf("%d ", HostInfo.nTasks);  /* Anzahl Tasks in MPI_COMM_WORLD */
	printf("%s ", HostInfo.Name);  /* Hostname des Rechners */
	printf("%s ", HostInfo.Machine); /* Maschinentyp z.B. sun4m */
	printf("%s ", HostInfo.OS);  /* Betriebssystem */
	printf("%s ", HostInfo.OSrelease); /* Versionsnummer des Betriebssystems */
	printf("%s ", HostInfo.OSversion); /* Kernelnummer bzw. Version */
	printf("%s \n", HostInfo.CPU);          /* CPU Typ */
	fflush(stdout);

	RankList = (int*) calloc(HostInfo.nTasks, sizeof(int));
	CPUIdList = (CPUType*) calloc(HostInfo.nTasks, sizeof(CPUType));
	NameList = (char**) malloc(HostInfo.nTasks * sizeof(char*));
	NameList[0] = (char*) calloc(HostInfo.nTasks * 80, sizeof(char));
	for (i = 1; i < HostInfo.nTasks; ++i) {
		NameList[i] = NameList[0] + (i * 80);
	}

	MPI_Allgather(HostInfo.Name, 80, MPI_CHAR,
	              NameList[0], 80, MPI_CHAR, MPI_COMM_WORLD);

	MPI_Allgather(&HostInfo.CPUID, 1, MPI_INT,
	              CPUIdList, 1, MPI_INT, MPI_COMM_WORLD);


	free(NameList[0]);
	free(NameList);
	free(CPUIdList);
	free(RankList);

	MPI_Finalize();
	return 0;
}
