#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <mpi.h>

#define MASTER 0

typedef enum ArchType_ {UNKNOWN, LINUX, KSR, SUN4, SOLARIS} ArchType;
typedef enum CPUType_ {OTHERCPU, PPRO, P2KLAMATH, P2DESCHUTES} CPUType;

typedef enum FamilyType_ {OTHERFAMILY, SIMPSON, BOUVIER} FamilyType;

char *ArchName[] = {"Unknown", "LINUX", "ksr", "sun4", "solaris"};
char *CPUName[] = {"Unknown", "Pentium Pro", "Pentium II (Klamath)", "Pentium II (Deschutes)"};

typedef struct HostInfo_
{
	int      Rank;
	int      nTasks;
	ArchType OSID;
	CPUType  CPUID;
	FamilyType Family;
	char     Name[80];
	char     Machine[80];
	char     Arch[80];
	char     OS[80];
	char     CPU[80];
	char     OSversion[80];
	char     OSrelease[80];
}
HostInfo_t;

char *SysCommand(const char *command)
{
	FILE *fp;
	static char Buffer[256];
	int i;

	i = 0;
	fp = popen(command, "r");
	Buffer[i] = fgetc(fp);
	if (Buffer[i] != '\n') ++i;
	while (!feof(fp)) {
		Buffer[i] = fgetc(fp);
		if (Buffer[i] != '\n') ++i;
	}
	Buffer[i-1] = 0x0;
	pclose(fp);
	return Buffer;
}

void getarch(HostInfo_t *HostInfo)
{
	FILE *fp;
	char FileName[80];
	char Arch[80], cpu[80];

	strcpy(FileName, "/bin/uname");
	if ((fp = fopen(FileName, "r")) != NULL) {
		fclose(fp);
		sprintf(Arch, "%s", SysCommand("/bin/uname -s"));
	}

	if ((fp = fopen("/bin/uname", "r")) != NULL) {
		fclose(fp);
		sprintf(HostInfo->OSrelease, "%s", SysCommand("/bin/uname -r"));
		sprintf(HostInfo->OSversion, "%s", SysCommand("/bin/uname -v"));
		sprintf(HostInfo->Machine, "%s", SysCommand("/bin/uname -m"));
	}

	gethostname(HostInfo->Name, 79);

	if (!strcmp(Arch, "Linux")) {
		HostInfo->OSID = LINUX;
	} else if (!strcmp(Arch, "ksr1")) {
		HostInfo->OSID = KSR;
	} else if (!strcmp(Arch, "sun4")) {
		HostInfo->OSID = SUN4;
	} else if (!strcmp(Arch, "SunOS")) {
		if (HostInfo->OSrelease[0] == '5')
			HostInfo->OSID = SOLARIS;
		else
			HostInfo->OSID = SUN4;
	}

	strcpy(HostInfo->OS, ArchName[HostInfo->OSID]);

	sprintf(cpu, "%s", SysCommand("cat /proc/cpuinfo | grep \"model name\"| head -n 1 | cut -f 2 -d :"));
	if (strstr(cpu, " Pentium Pro")) {
		HostInfo->CPUID = PPRO;
	} else if (strstr(cpu, " Pentium II (Klamath)")) {
		HostInfo->CPUID = P2KLAMATH;
	} else if (strstr(cpu, " Pentium II (Deschutes)")) {
		HostInfo->CPUID = P2DESCHUTES;
	} else {
		HostInfo->CPUID = OTHERCPU;
	}

	if ((strcmp("homer", HostInfo->Name) == 0) ||
	                (strcmp("marge", HostInfo->Name) == 0) ||
	                (strcmp("lisa", HostInfo->Name) == 0)) {
		HostInfo->Family = SIMPSON;
	} else if ((strcmp("patty", HostInfo->Name) == 0) ||
	                (strcmp("selma", HostInfo->Name) == 0)) {
		HostInfo->Family = BOUVIER;
	} else {
		HostInfo->Family = OTHERFAMILY;
	}

	strcpy(HostInfo->CPU, CPUName[HostInfo->CPUID]);

	MPI_Comm_rank(MPI_COMM_WORLD, &HostInfo->Rank);
	MPI_Comm_size(MPI_COMM_WORLD, &HostInfo->nTasks);
}

int main(int argc, char *argv[])
{
	int  *RankList;
	CPUType         *CPUIdList;
	FamilyType      *FamilyList;
	char  **NameList;
	int  i;
	HostInfo_t HostInfo;
	int NewGroupSize;
	MPI_Group NewGroup;
	MPI_Group OtherGroup;
	MPI_Group WorldGroup;
	MPI_Comm OtherComm;
	MPI_Comm NewComm;
	char NewMaster[80];
	int NewRank;
	int NewSize;
	int OtherSize;

	MPI_Init(&argc, &argv);
	getarch(&HostInfo);

	RankList = (int*) calloc(HostInfo.nTasks, sizeof(int));
	CPUIdList = (CPUType*) calloc(HostInfo.nTasks, sizeof(CPUType));
	FamilyList = (FamilyType*) calloc(HostInfo.nTasks, sizeof(FamilyType));
	NameList = (char**) malloc(HostInfo.nTasks * sizeof(char*));
	NameList[0] = (char*) calloc(HostInfo.nTasks * 80, sizeof(char));
	for (i = 1; i < HostInfo.nTasks; ++i) {
		NameList[i] = NameList[0] + (i * 80);
	}

	MPI_Allgather(HostInfo.Name, 80, MPI_CHAR,
	              NameList[0], 80, MPI_CHAR, MPI_COMM_WORLD);

	MPI_Allgather(&HostInfo.CPUID, 1, MPI_INT,
	              CPUIdList, 1, MPI_INT, MPI_COMM_WORLD);

	MPI_Allgather(&HostInfo.Family, 1, MPI_INT,
	              FamilyList, 1, MPI_INT, MPI_COMM_WORLD);

	NewGroupSize = 0;
	for (i = 0; i < HostInfo.nTasks; i++) {
		if (FamilyList[i] == FamilyList[HostInfo.Rank]) {
			RankList[NewGroupSize] = i;
			NewGroupSize++;
		}
	}

	MPI_Comm_group(MPI_COMM_WORLD, &WorldGroup);
	MPI_Group_incl(WorldGroup, NewGroupSize, RankList, &NewGroup);
	MPI_Group_excl(WorldGroup, NewGroupSize, RankList, &OtherGroup);
	MPI_Comm_create(MPI_COMM_WORLD, NewGroup, &NewComm);
	MPI_Comm_create(MPI_COMM_WORLD, OtherGroup, &OtherComm);

	strcpy(NewMaster, HostInfo.Name);
	MPI_Bcast(NewMaster, 80, MPI_CHAR, 0 , NewComm);

	MPI_Comm_rank(NewComm, &NewRank);
	MPI_Comm_size(NewComm, &NewSize);
	MPI_Comm_size(OtherComm, &OtherSize);

	printf("%s WORLD(%d/%d) New(%d/%d) new Slave of %s - misses %d nodes\n", HostInfo.Name,
	       HostInfo.Rank, HostInfo.nTasks, NewRank, NewSize, NewMaster, OtherSize);

	MPI_Comm_free(&NewComm);
	MPI_Group_free(&NewGroup);
	MPI_Group_free(&WorldGroup);

	free(NameList[0]);
	free(NameList);
	free(FamilyList);
	free(CPUIdList);
	free(RankList);

	MPI_Finalize();
	return 0;
}
